/*****************************************************************************
 *
 *                        CalC (Calcium Calculator)
 *                             Victor Matveev
 *
 *                                vector.cpp
 *
 *                   A basic linear algebra vector class 
 *
 ******************************************************************************/
/*
 * 2/19/01: fixed bug in norm_diff(): using the "tiny" constants screwes up the
 *          continuity of norm calculation
 *
 */

#include "stdafx.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>  // for compatibility with Visual C++
#include <string.h>
#include "vector.h"

extern int xsize, ysize, zsize;
extern FILE *stdlog;

extern FILE *fopenAssure(const char *fname, const char *mode, const char *action, const char *id);
extern char *makeMessage(const char *fmt, ...);

//*******************************************************************************
//                       I M P L E M E N T A T I O N
//*******************************************************************************


VectorObj::VectorObj(const VectorObj &v)
{
  //  fprintf(stderr,"copy constructor for VectorObj with %d elements\n", v.size);
  //  fprintf(stderr,"pass 1: size=%ld\n",size);
  size = v.size;

  elem = new double[size];
  for (long i = 0; i < size; ++i)  elem[i] = v.elem[i];
}

//==============================================================================

VectorObj &VectorObj::operator=(const double *e)
{
	for (long i = 0; i < size; ++i) elem[i] = e[i];
	return *this;
}

VectorObj &VectorObj::operator=(const double val)
{
	for (long i = 0; i < size; ++i) elem[i] = val;
	return *this;
}

VectorObj &VectorObj::operator=(const VectorObj &v)
{
	for (long i = 0; i < size; ++i) elem[i] = v.elem[i];
	return *this;
}


//------------------------------------------------------------------------------


VectorObj *vabs(const VectorObj &v)
{
       VectorObj *vv = new VectorObj(v.size);
       for (long i = 0; i < v.size; ++i) (*vv).elem[i] = fabs(v.elem[i]);
       return vv;
}

//------------------------------------------------------------------------------


VectorObj operator-(const VectorObj &v)
{
       VectorObj vv(v.size);
       for (long i = 0; i < v.size; ++i) vv.elem[i] = -v.elem[i];
       return vv;
}


VectorObj operator-(const VectorObj &v1, const VectorObj &v2)
{
       VectorObj v3(v1.size);
       for (long i = 0; i < v1.size; ++i)  v3.elem[i] = v1.elem[i]-v2.elem[i];
       return v3;
}

VectorObj operator-(const VectorObj &v, const double val)
{
       VectorObj vv(v.size);
       for (long i = 0; i < v.size; ++i)  vv.elem[i] = v.elem[i] - val;
       return vv;
}

VectorObj operator-(const double val, const VectorObj &v)
{
       VectorObj vv(v.size);
       for (long i = 0; i < v.size; ++i)  vv.elem[i] = val - v.elem[i];
       return vv;
}

void operator-=(VectorObj &v1, const VectorObj &v2)
{
       for (long i = 0; i < v1.size; ++i)  v1.elem[i] -= v2.elem[i];
}


void operator-=(VectorObj &v, const double val)
{
       for (long i = 0; i < v.size; ++i)  v.elem[i] -= val;
}

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


VectorObj operator+(const VectorObj &v1, const VectorObj &v2)
{
       VectorObj v3(v1.size);
       for (long i = 0; i < v1.size; ++i)  v3.elem[i] = v1.elem[i] + v2.elem[i];
       return v3;
}

VectorObj operator+(const VectorObj &v, const double val)
{
       VectorObj vv(v.size);
       for (long i = 0; i < v.size; ++i)  vv.elem[i] = v.elem[i] + val;
       return vv;
}

VectorObj operator+(const double val, const VectorObj &v)
{
       VectorObj vv(v.size);
       for (long i = 0; i < v.size; ++i)  vv.elem[i] = v.elem[i] + val;
       return vv;
}

void operator+=(VectorObj &v1, const VectorObj &v2)
{
       for (long i = 0; i < v1.size; ++i)  v1.elem[i] += v2.elem[i];
       //       return v1;
}

void operator+=(VectorObj &v, const double val)
{
       for (long i = 0; i < v.size; ++i)  v.elem[i] += val;
       //return v;
}

// inline VectorObj& operator+(VectorObj &v)
// {
// return v;
// }

//*******************************************************************************

VectorObj operator%(const VectorObj &v1, const VectorObj &v2)
{
       VectorObj v3(v1.size);
       for (long i = 0; i < v1.size; ++i)  v3.elem[i] = v1.elem[i] * v2.elem[i];
       return v3;
}


double operator*(const VectorObj &v1, const VectorObj &v2)
{
       double sum = 0.0;
       for (long i = 0; i < v1.size; ++i)  sum += v1.elem[i] * v2.elem[i];
       return sum;
}

VectorObj operator*(const VectorObj &v, const double val)
{
       VectorObj vv(v.size);
       for (long i = 0; i < v.size; ++i)  vv.elem[i] = v.elem[i] * val;
       return vv;
}

VectorObj operator*(const double val, const VectorObj &v)
{
       VectorObj vv(v.size);
       for (long i = 0; i < v.size; ++i)  vv.elem[i] = v.elem[i] * val;
       return vv;
}

VectorObj operator%=(VectorObj &v1, const VectorObj &v2)
{
       for (long i = 0; i < v1.size; ++i)  v1.elem[i] *= v2.elem[i];
       return v1;
}


VectorObj operator*=(VectorObj &v, const double val)
{
       for (long i = 0; i < v.size; ++i)  v.elem[i] *= val;
       return v;
}

/****************************************************************************/

void  VectorObj::print(int n) const
{
        char s1[20], s2[20];

        strcpy(s1,"%.");
        sprintf(s2,"%d",n);
        strcat(s1,s2);
        strcat(s1,"g ");

	for (long i = 0; i < size; ++i)  fprintf(stdlog, s1, elem[i]);
	printf("\n");
}
/*
double  VectorObj::pnorm() const
{
        double sum = 0.0;

	for (long i = 0; i < size; ++i) 
          if (elem[i] >= 0) sum += elem[i]*elem[i]; else return -1;
	return sqrt(sum);
}
*/
double  VectorObj::norm() const
{
        double sum = 0.0;

	for (long i = 0; i < size; ++i) sum += elem[i]*elem[i]; 
	return sqrt(sum);
}

double  VectorObj::collapse() const
{
        double sum = 0.0;

	for (long i = 0; i < size; ++i) sum += elem[i];
	return sum;
}

//****************************************************************************/
/*
double  VectorObj::fourier111() const
{
  int  ix = 0, iy = 0, iz = 0;
  long i;
  double sum = 0.0;
  bool s = false;

  for (i = 0; i < size; ++i)
     {
     if (s = !s) sum += elem[i];
        else sum -= elem[i];

     if (++ix == xsize)
	 {
	   ix = 0; 
	   if (++iy == ysize)
	     {
	       iy = 0;
	       if (++ iz == zsize) iz = 0;
	     }
	 }
     }

  //  fprintf(stderr,"fourier111 = %g\n",sum);
 return sum;
}
*/
//****************************************************************************/

double stiff_norm_diff(VectorObj &dv, VectorObj &v)
{
   double maxerr = 0.0, temp = 0.0;

   for (long i = 0; i < v.size; ++i)
     { 
     temp = fabs(dv[i] / v[i]);
     if ( isnan(temp) ) return temp;
     if (temp > maxerr) maxerr = temp;
     }

   return maxerr;
}

//****************************************************************************/

double stiff_norm(VectorObj &dv)
{
   double maxerr = 0.0, temp = 0.0;

   for (long i = 0; i < dv.size; ++i)
     { 
     temp = fabs(dv[i]);
     if ( isnan(temp) ) return temp;
     if (temp > maxerr) maxerr = temp;
     }

   return maxerr;
}
//****************************************************************************/

double norm_diff(VectorObj &v1, VectorObj &v2)
{
   double sum1 = 0.0, sum2 = 0.0;

   for (long i = 0; i < v1.size; ++i)
     { 
     sum1 += (v1[i] - v2[i]) * (v1[i] - v2[i]);
     sum2 += (v1[i] + v2[i]) * (v1[i] + v2[i]);
     }

   if (sum2 > 0) return sqrt(sum1/sum2); 
   else return 0.0;
}

//****************************************************************************/

void VectorObj::Export(const char *filename)
{
  FILE *f = fopenAssure(filename, "wb", "Vector export", "");

  fwrite( (void *)(&size), sizeof(long),   1, f);
  fwrite( (void *)elem,    sizeof(double), size, f);

  fclose(f);

}

//****************************************************************************/

void VectorObj::import(const char *filename)
{
  FILE *f = fopenAssure(filename, "rb", "Vector import", "");

  long n;  
  fread( (void *)(&n), sizeof(long),   1, f);

  if (n != size) throw makeMessage("in Vector import: wrong number of elements (%ld vs %ld)\n",n,size);

  fread( (void *)elem, sizeof(double), size, f);

  fclose(f);

}

//****************************************************************************/
